# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import numpy
import pytest

import modin.numpy as np

from .utils import assert_scalar_or_array_equal


def test_max():
    # Test 1D
    numpy_arr = numpy.random.randint(-100, 100, size=100)
    modin_arr = np.array(numpy_arr)
    assert modin_arr.max() == numpy_arr.max()
    modin_result = modin_arr.max(axis=0)
    numpy_result = modin_arr.max(axis=0)
    assert modin_result == numpy_result
    modin_result = modin_arr.max(initial=200)
    numpy_result = numpy_arr.max(initial=200)
    assert modin_result == numpy_result
    modin_result = modin_arr.max(initial=0, where=False)
    numpy_result = numpy_arr.max(initial=0, where=False)
    assert modin_result == numpy_result
    modin_result = modin_arr.max(keepdims=True)
    numpy_result = numpy_arr.max(keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_arr = numpy.array([1, 10000, 2, 3, 4, 5])
    modin_arr = np.array(numpy_arr)
    numpy_mask = numpy.array([True, False, True, True, True, True])
    modin_mask = np.array(numpy_mask)
    assert numpy_arr.max(where=numpy_mask, initial=5) == modin_arr.max(
        where=modin_mask, initial=5
    )
    # Test 2D
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    assert modin_arr.max() == numpy_arr.max()
    modin_result = modin_arr.max(axis=0)
    numpy_result = numpy_arr.max(axis=0)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.max(axis=0, keepdims=True)
    numpy_result = numpy_arr.max(axis=0, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.max(axis=1)
    numpy_result = numpy_arr.max(axis=1)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.max(axis=1, keepdims=True)
    numpy_result = numpy_arr.max(axis=1, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.max(initial=200)
    numpy_result = numpy_arr.max(initial=200)
    assert modin_result == numpy_result
    modin_result = modin_arr.max(initial=0, where=False)
    numpy_result = numpy_arr.max(initial=0, where=False)
    assert modin_result == numpy_result
    with pytest.raises(ValueError):
        modin_arr.max(out=modin_arr, keepdims=True)
    modin_out = np.array([[1]])
    numpy_out = modin_out._to_numpy()
    modin_result = modin_arr.max(out=modin_out, keepdims=True)
    numpy_result = numpy_arr.max(out=numpy_out, keepdims=True)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    modin_result = modin_arr.max(axis=0, where=False, initial=4)
    numpy_result = numpy_arr.max(axis=0, where=False, initial=4)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.max(axis=0, where=False, initial=4, out=modin_out)
    numpy_result = numpy_arr.max(axis=0, where=False, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.max(axis=0, initial=4, out=modin_out)
    numpy_result = numpy_arr.max(axis=0, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.max(axis=1, initial=4, out=modin_out)
    numpy_result = numpy_arr.max(axis=1, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    numpy_where = numpy.full(20, False)
    numpy_where[:10] = True
    numpy.random.shuffle(numpy_where)
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.max(axis=0, initial=4, out=modin_out, where=modin_where)
    numpy_result = numpy_arr.max(axis=0, initial=4, out=numpy_out, where=numpy_where)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.array([[1, 10000, 2], [3, 4, 5]])
    modin_arr = np.array(numpy_arr)
    numpy_mask = numpy.array([[True, False, True], [True, True, True]])
    modin_mask = np.array(numpy_mask)
    assert_scalar_or_array_equal(
        modin_arr.max(where=modin_mask, initial=5),
        numpy_arr.max(where=numpy_mask, initial=5),
    )


def test_min():
    # Test 1D
    numpy_arr = numpy.random.randint(-100, 100, size=100)
    modin_arr = np.array(numpy_arr)
    assert modin_arr.min() == numpy_arr.min()
    modin_result = modin_arr.min(axis=0)
    numpy_result = modin_arr.min(axis=0)
    assert modin_result == numpy_result
    modin_result = modin_arr.min(initial=-200)
    numpy_result = numpy_arr.min(initial=-200)
    assert modin_result == numpy_result
    modin_result = modin_arr.min(initial=0, where=False)
    numpy_result = numpy_arr.min(initial=0, where=False)
    assert modin_result == numpy_result
    modin_result = modin_arr.min(keepdims=True)
    numpy_result = numpy_arr.min(keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_arr = numpy.array([1, -10000, 2, 3, 4, 5])
    modin_arr = np.array(numpy_arr)
    numpy_mask = numpy.array([True, False, True, True, True, True])
    modin_mask = np.array(numpy_mask)
    assert numpy_arr.min(where=numpy_mask, initial=5) == modin_arr.min(
        where=modin_mask, initial=5
    )
    # Test 2D
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    assert modin_arr.min() == numpy_arr.min()
    modin_result = modin_arr.min(axis=0)
    numpy_result = numpy_arr.min(axis=0)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.min(axis=0, keepdims=True)
    numpy_result = numpy_arr.min(axis=0, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.min(axis=1)
    numpy_result = numpy_arr.min(axis=1)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.min(axis=1, keepdims=True)
    numpy_result = numpy_arr.min(axis=1, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.min(initial=-200)
    numpy_result = numpy_arr.min(initial=-200)
    assert modin_result == numpy_result
    modin_result = modin_arr.min(initial=0, where=False)
    numpy_result = numpy_arr.min(initial=0, where=False)
    assert modin_result == numpy_result
    with pytest.raises(ValueError):
        modin_arr.min(out=modin_arr, keepdims=True)
    modin_out = np.array([[1]])
    numpy_out = modin_out._to_numpy()
    modin_result = modin_arr.min(out=modin_out, keepdims=True)
    numpy_result = numpy_arr.min(out=numpy_out, keepdims=True)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    modin_result = modin_arr.min(axis=0, where=False, initial=4)
    numpy_result = numpy_arr.min(axis=0, where=False, initial=4)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.min(axis=0, where=False, initial=4, out=modin_out)
    numpy_result = numpy_arr.min(axis=0, where=False, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.min(axis=0, initial=4, out=modin_out)
    numpy_result = numpy_arr.min(axis=0, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.min(axis=1, initial=4, out=modin_out)
    numpy_result = numpy_arr.min(axis=1, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    numpy_where = numpy.full(20, False)
    numpy_where[:10] = True
    numpy.random.shuffle(numpy_where)
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.min(axis=0, initial=4, out=modin_out, where=modin_where)
    numpy_result = numpy_arr.min(axis=0, initial=4, out=numpy_out, where=numpy_where)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.array([[1, -10000, 2], [3, 4, 5]])
    modin_arr = np.array(numpy_arr)
    numpy_mask = numpy.array([[True, False, True], [True, True, True]])
    modin_mask = np.array(numpy_mask)
    assert_scalar_or_array_equal(
        modin_arr.min(where=modin_mask, initial=5),
        numpy_arr.min(where=numpy_mask, initial=5),
    )


def test_sum():
    # Test 1D
    numpy_arr = numpy.random.randint(-100, 100, size=100)
    modin_arr = np.array(numpy_arr)
    assert modin_arr.sum() == numpy_arr.sum()
    modin_result = modin_arr.sum(axis=0)
    numpy_result = modin_arr.sum(axis=0)
    assert modin_result == numpy_result
    modin_result = modin_arr.sum(initial=-200)
    numpy_result = numpy_arr.sum(initial=-200)
    assert modin_result == numpy_result
    modin_result = modin_arr.sum(initial=0, where=False)
    numpy_result = numpy_arr.sum(initial=0, where=False)
    assert modin_result == numpy_result
    modin_result = modin_arr.sum(keepdims=True)
    numpy_result = numpy_arr.sum(keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_arr = numpy.array([1, 10000, 2, 3, 4, 5])
    modin_arr = np.array(numpy_arr)
    numpy_mask = numpy.array([True, False, True, True, True, True])
    modin_mask = np.array(numpy_mask)
    assert numpy_arr.sum(where=numpy_mask) == modin_arr.sum(where=modin_mask)
    # Test 2D
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    assert modin_arr.sum() == numpy_arr.sum()
    modin_result = modin_arr.sum(axis=0)
    numpy_result = numpy_arr.sum(axis=0)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.sum(axis=0, keepdims=True)
    numpy_result = numpy_arr.sum(axis=0, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.sum(axis=1)
    numpy_result = numpy_arr.sum(axis=1)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.sum(axis=1, keepdims=True)
    numpy_result = numpy_arr.sum(axis=1, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.sum(initial=-200)
    numpy_result = numpy_arr.sum(initial=-200)
    assert modin_result == numpy_result
    modin_result = modin_arr.sum(initial=0, where=False)
    numpy_result = numpy_arr.sum(initial=0, where=False)
    assert modin_result == numpy_result
    with pytest.raises(ValueError):
        modin_arr.sum(out=modin_arr, keepdims=True)
    modin_out = np.array([[1]])
    numpy_out = modin_out._to_numpy()
    modin_result = modin_arr.sum(out=modin_out, keepdims=True)
    numpy_result = numpy_arr.sum(out=numpy_out, keepdims=True)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    modin_result = modin_arr.sum(axis=0, where=False, initial=4)
    numpy_result = numpy_arr.sum(axis=0, where=False, initial=4)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.sum(axis=0, where=False, initial=4, out=modin_out)
    numpy_result = numpy_arr.sum(axis=0, where=False, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.sum(axis=0, initial=4, out=modin_out)
    numpy_result = numpy_arr.sum(axis=0, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.sum(axis=1, initial=4, out=modin_out)
    numpy_result = numpy_arr.sum(axis=1, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    numpy_where = numpy.full(20, False)
    numpy_where[:10] = True
    numpy.random.shuffle(numpy_where)
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.sum(axis=0, initial=4, out=modin_out, where=modin_where)
    numpy_result = numpy_arr.sum(axis=0, initial=4, out=numpy_out, where=numpy_where)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_where = numpy.full(400, False)
    numpy_where[:200] = True
    numpy.random.shuffle(numpy_where)
    numpy_where = numpy_where.reshape((20, 20))
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.sum(where=modin_where)
    numpy_result = numpy_arr.sum(where=numpy_where)
    assert modin_result == numpy_result
    # Test NA propagation
    numpy_arr = numpy.array([[1, 2], [3, 4], [5, numpy.nan]])
    modin_arr = np.array([[1, 2], [3, 4], [5, np.nan]])
    assert numpy.isnan(modin_arr.sum())
    assert_scalar_or_array_equal(
        modin_arr.sum(axis=1),
        numpy_arr.sum(axis=1),
    )
    assert_scalar_or_array_equal(
        modin_arr.sum(axis=0),
        numpy_arr.sum(axis=0),
    )


def test_mean():
    # Test 1D
    numpy_arr = numpy.random.randint(-100, 100, size=100)
    modin_arr = np.array(numpy_arr)
    assert modin_arr.mean() == numpy_arr.mean()
    modin_result = modin_arr.mean(axis=0)
    numpy_result = modin_arr.mean(axis=0)
    assert modin_result == numpy_result
    modin_result = modin_arr.mean()
    numpy_result = numpy_arr.mean()
    assert modin_result == numpy_result
    modin_result = modin_arr.mean(keepdims=True)
    numpy_result = numpy_arr.mean(keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_arr = numpy.array([1, 10000, 2, 3, 4, 5])
    modin_arr = np.array(numpy_arr)
    numpy_mask = numpy.array([True, False, True, True, True, True])
    modin_mask = np.array(numpy_mask)
    assert numpy_arr.mean(where=numpy_mask) == modin_arr.mean(where=modin_mask)
    # Test 2D
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    assert modin_arr.mean() == numpy_arr.mean()
    modin_result = modin_arr.mean(axis=0)
    numpy_result = numpy_arr.mean(axis=0)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.mean(axis=0, keepdims=True)
    numpy_result = numpy_arr.mean(axis=0, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.mean(axis=1)
    numpy_result = numpy_arr.mean(axis=1)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.mean(axis=1, keepdims=True)
    numpy_result = numpy_arr.mean(axis=1, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.mean()
    numpy_result = numpy_arr.mean()
    assert modin_result == numpy_result
    with pytest.raises(ValueError):
        modin_arr.mean(out=modin_arr, keepdims=True)
    modin_out = np.array([[1]])
    numpy_out = modin_out._to_numpy()
    modin_result = modin_arr.mean(out=modin_out, keepdims=True)
    numpy_result = numpy_arr.mean(out=numpy_out, keepdims=True)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.mean(axis=0, where=False, out=modin_out)
    numpy_result = numpy_arr.mean(axis=0, where=False, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.mean(axis=0, out=modin_out)
    numpy_result = numpy_arr.mean(axis=0, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.mean(axis=1, out=modin_out)
    numpy_result = numpy_arr.mean(axis=1, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    numpy_where = numpy.full(20, False)
    numpy_where[:10] = True
    numpy.random.shuffle(numpy_where)
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.mean(axis=0, out=modin_out, where=modin_where)
    numpy_result = numpy_arr.mean(axis=0, out=numpy_out, where=numpy_where)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_where = numpy.full(400, False)
    numpy_where[:200] = True
    numpy.random.shuffle(numpy_where)
    numpy_where = numpy_where.reshape((20, 20))
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.mean(where=modin_where)
    numpy_result = numpy_arr.mean(where=numpy_where)
    assert modin_result == numpy_result
    # Test NA propagation
    numpy_arr = numpy.array([[1, 2], [3, 4], [5, numpy.nan]])
    modin_arr = np.array([[1, 2], [3, 4], [5, np.nan]])
    assert numpy.isnan(modin_arr.mean())
    assert_scalar_or_array_equal(
        modin_arr.mean(axis=1),
        numpy_arr.mean(axis=1),
    )
    assert_scalar_or_array_equal(
        modin_arr.mean(axis=0),
        numpy_arr.mean(axis=0),
    )
    numpy_where = numpy.array([[True, True], [True, True], [True, False]])
    modin_where = np.array(numpy_where)
    assert modin_arr.mean(where=modin_where) == numpy_arr.mean(where=numpy_where)


def test_prod():
    # Test 1D
    numpy_arr = numpy.random.randint(-100, 100, size=100)
    modin_arr = np.array(numpy_arr)
    assert modin_arr.prod() == numpy_arr.prod()
    modin_result = modin_arr.prod(axis=0)
    numpy_result = modin_arr.prod(axis=0)
    assert modin_result == numpy_result
    modin_result = modin_arr.prod(initial=-200)
    numpy_result = numpy_arr.prod(initial=-200)
    assert modin_result == numpy_result
    modin_result = modin_arr.prod(initial=0, where=False)
    numpy_result = numpy_arr.prod(initial=0, where=False)
    assert modin_result == numpy_result
    modin_result = modin_arr.prod(keepdims=True)
    numpy_result = numpy_arr.prod(keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_arr = numpy.array([1, 10000, 2, 3, 4, 5])
    modin_arr = np.array(numpy_arr)
    numpy_mask = numpy.array([True, False, True, True, True, True])
    modin_mask = np.array(numpy_mask)
    assert numpy_arr.prod(where=numpy_mask) == modin_arr.prod(where=modin_mask)
    # Test 2D
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    assert modin_arr.prod() == numpy_arr.prod()
    modin_result = modin_arr.prod(axis=0)
    numpy_result = numpy_arr.prod(axis=0)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.prod(axis=0, keepdims=True)
    numpy_result = numpy_arr.prod(axis=0, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.prod(axis=1)
    numpy_result = numpy_arr.prod(axis=1)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.prod(axis=1, keepdims=True)
    numpy_result = numpy_arr.prod(axis=1, keepdims=True)
    assert modin_result.shape == numpy_result.shape
    assert_scalar_or_array_equal(modin_result, numpy_result)
    modin_result = modin_arr.prod(initial=-200)
    numpy_result = numpy_arr.prod(initial=-200)
    assert modin_result == numpy_result
    modin_result = modin_arr.prod(initial=0, where=False)
    numpy_result = numpy_arr.prod(initial=0, where=False)
    assert modin_result == numpy_result
    with pytest.raises(ValueError):
        modin_arr.prod(out=modin_arr, keepdims=True)
    modin_out = np.array([[1]])
    numpy_out = modin_out._to_numpy()
    modin_result = modin_arr.prod(out=modin_out, keepdims=True)
    numpy_result = numpy_arr.prod(out=numpy_out, keepdims=True)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    modin_result = modin_arr.prod(axis=0, where=False, initial=4)
    numpy_result = numpy_arr.prod(axis=0, where=False, initial=4)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.prod(axis=0, where=False, initial=4, out=modin_out)
    numpy_result = numpy_arr.prod(axis=0, where=False, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_arr = numpy.random.randint(-100, 100, size=(20, 20))
    modin_arr = np.array(numpy_arr)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.prod(axis=0, initial=4, out=modin_out)
    numpy_result = numpy_arr.prod(axis=0, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    modin_result = modin_arr.prod(axis=1, initial=4, out=modin_out)
    numpy_result = numpy_arr.prod(axis=1, initial=4, out=numpy_out)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_out = numpy.ones(20)
    modin_out = np.array(numpy_out)
    numpy_where = numpy.full(20, False)
    numpy_where[:10] = True
    numpy.random.shuffle(numpy_where)
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.prod(axis=0, initial=4, out=modin_out, where=modin_where)
    numpy_result = numpy_arr.prod(axis=0, initial=4, out=numpy_out, where=numpy_where)
    assert_scalar_or_array_equal(modin_result, numpy_result)
    assert_scalar_or_array_equal(modin_out, numpy_out)
    numpy_where = numpy.full(400, False)
    numpy_where[:200] = True
    numpy.random.shuffle(numpy_where)
    numpy_where = numpy_where.reshape((20, 20))
    modin_where = np.array(numpy_where)
    modin_result = modin_arr.prod(where=modin_where)
    numpy_result = numpy_arr.prod(where=numpy_where)
    assert modin_result == numpy_result
    # Test NA propagation
    numpy_arr = numpy.array([[1, 2], [3, 4], [5, numpy.nan]])
    modin_arr = np.array([[1, 2], [3, 4], [5, np.nan]])
    assert numpy.isnan(modin_arr.prod())
    assert_scalar_or_array_equal(
        modin_arr.prod(axis=1),
        numpy_arr.prod(axis=1),
    )
    assert_scalar_or_array_equal(
        modin_arr.prod(axis=0),
        numpy_arr.prod(axis=0),
    )



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""
Module houses `BaseIO` class.

`BaseIO` is base class for IO classes, that stores IO functions.
"""

from typing import Any

import pandas
from pandas._libs.lib import no_default
from pandas.util._decorators import doc

from modin.core.storage_formats.base.query_compiler import BaseQueryCompiler
from modin.db_conn import ModinDatabaseConnection
from modin.error_message import ErrorMessage
from modin.pandas.io import ExcelFile
from modin.utils import _inherit_docstrings

_doc_default_io_method = """
{summary} using pandas.
For parameters description please refer to pandas API.

Returns
-------
{returns}
"""

_doc_returns_qc = """BaseQueryCompiler
    QueryCompiler with read data."""

_doc_returns_qc_or_parser = """BaseQueryCompiler or TextParser
    QueryCompiler or TextParser with read data."""


class BaseIO:
    """Class for basic utils and default implementation of IO functions."""

    query_compiler_cls: BaseQueryCompiler = None
    frame_cls = None

    @classmethod
    def from_non_pandas(cls, *args, **kwargs):
        """
        Create a Modin `query_compiler` from a non-pandas `object`.

        Parameters
        ----------
        *args : iterable
            Positional arguments to be passed into `func`.
        **kwargs : dict
            Keyword arguments to be passed into `func`.
        """
        return None

    @classmethod
    def from_pandas(cls, df):
        """
        Create a Modin `query_compiler` from a `pandas.DataFrame`.

        Parameters
        ----------
        df : pandas.DataFrame
            The pandas DataFrame to convert from.

        Returns
        -------
        BaseQueryCompiler
            QueryCompiler containing data from the `pandas.DataFrame`.
        """
        return cls.query_compiler_cls.from_pandas(df, cls.frame_cls)

    @classmethod
    def from_arrow(cls, at):
        """
        Create a Modin `query_compiler` from a `pyarrow.Table`.

        Parameters
        ----------
        at : Arrow Table
            The Arrow Table to convert from.

        Returns
        -------
        BaseQueryCompiler
            QueryCompiler containing data from the Arrow Table.
        """
        return cls.query_compiler_cls.from_arrow(at, cls.frame_cls)

    @classmethod
    def from_dataframe(cls, df):
        """
        Create a Modin QueryCompiler from a DataFrame supporting the DataFrame exchange protocol `__dataframe__()`.

        Parameters
        ----------
        df : DataFrame
            The DataFrame object supporting the DataFrame exchange protocol.

        Returns
        -------
        BaseQueryCompiler
            QueryCompiler containing data from the DataFrame.
        """
        return cls.query_compiler_cls.from_dataframe(df, cls.frame_cls)

    @classmethod
    def from_ray_dataset(cls, ray_obj):
        """
        Create a Modin `query_compiler` from a Ray Dataset.

        Parameters
        ----------
        ray_obj : ray.data.Dataset
            The Ray Dataset to convert from.

        Returns
        -------
        BaseQueryCompiler
            QueryCompiler containing data from the Ray Dataset.

        Notes
        -----
        Ray Dataset can only be converted to a Modin Dataframe if Modin uses a Ray engine.
        If another engine is used, the runtime exception will be raised.
        """
        raise RuntimeError(
            "Modin Dataframe can only be converted to a Ray Dataset if Modin uses a Ray engine."
        )

    @classmethod
    @_inherit_docstrings(pandas.read_parquet, apilink="pandas.read_parquet")
    @doc(
        _doc_default_io_method,
        summary="Load a parquet object from the file path, returning a query compiler",
        returns=_doc_returns_qc,
    )
    def read_parquet(cls, **kwargs):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_parquet`")
        return cls.from_pandas(pandas.read_parquet(**kwargs))

    @classmethod
    @_inherit_docstrings(pandas.read_csv, apilink="pandas.read_csv")
    @doc(
        _doc_default_io_method,
        summary="Read a comma-separated values (CSV) file into query compiler",
        returns=_doc_returns_qc_or_parser,
    )
    def read_csv(
        cls,
        filepath_or_buffer,
        **kwargs,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_csv`")
        pd_obj = pandas.read_csv(filepath_or_buffer, **kwargs)
        if isinstance(pd_obj, pandas.DataFrame):
            return cls.from_pandas(pd_obj)
        if isinstance(pd_obj, pandas.io.parsers.TextFileReader):
            # Overwriting the read method should return a Modin DataFrame for calls
            # to __next__ and get_chunk
            pd_read = pd_obj.read
            pd_obj.read = lambda *args, **kw: cls.from_pandas(pd_read(*args, **kw))
        return pd_obj

    @classmethod
    @_inherit_docstrings(pandas.read_json, apilink="pandas.read_json")
    @doc(
        _doc_default_io_method,
        summary="Convert a JSON string to query compiler",
        returns=_doc_returns_qc,
    )
    def read_json(
        cls,
        **kwargs,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_json`")
        return cls.from_pandas(pandas.read_json(**kwargs))

    @classmethod
    @_inherit_docstrings(pandas.read_gbq, apilink="pandas.read_gbq")
    @doc(
        _doc_default_io_method,
        summary="Load data from Google BigQuery into query compiler",
        returns=_doc_returns_qc,
    )
    def read_gbq(
        cls,
        query: str,
        project_id=None,
        index_col=None,
        col_order=None,
        reauth=False,
        auth_local_webserver=False,
        dialect=None,
        location=None,
        configuration=None,
        credentials=None,
        use_bqstorage_api=None,
        private_key=None,
        verbose=None,
        progress_bar_type=None,
        max_results=None,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_gbq`")
        return cls.from_pandas(
            pandas.read_gbq(
                query,
                project_id=project_id,
                index_col=index_col,
                col_order=col_order,
                reauth=reauth,
                auth_local_webserver=auth_local_webserver,
                dialect=dialect,
                location=location,
                configuration=configuration,
                credentials=credentials,
                use_bqstorage_api=use_bqstorage_api,
                progress_bar_type=progress_bar_type,
                max_results=max_results,
            )
        )

    @classmethod
    @_inherit_docstrings(pandas.read_html, apilink="pandas.read_html")
    @doc(
        _doc_default_io_method,
        summary="Read HTML tables into query compiler",
        returns=_doc_returns_qc,
    )
    def read_html(
        cls,
        io,
        *,
        match=".+",
        flavor=None,
        header=None,
        index_col=None,
        skiprows=None,
        attrs=None,
        parse_dates=False,
        thousands=",",
        encoding=None,
        decimal=".",
        converters=None,
        na_values=None,
        keep_default_na=True,
        displayed_only=True,
        **kwargs,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_html`")
        result = pandas.read_html(
            io=io,
            match=match,
            flavor=flavor,
            header=header,
            index_col=index_col,
            skiprows=skiprows,
            attrs=attrs,
            parse_dates=parse_dates,
            thousands=thousands,
            encoding=encoding,
            decimal=decimal,
            converters=converters,
            na_values=na_values,
            keep_default_na=keep_default_na,
            displayed_only=displayed_only,
            **kwargs,
        )
        return (cls.from_pandas(df) for df in result)

    @classmethod
    @_inherit_docstrings(pandas.read_clipboard, apilink="pandas.read_clipboard")
    @doc(
        _doc_default_io_method,
        summary="Read text from clipboard into query compiler",
        returns=_doc_returns_qc,
    )
    def read_clipboard(cls, sep=r"\s+", **kwargs):  # pragma: no cover # noqa: PR01
        ErrorMessage.default_to_pandas("`read_clipboard`")
        return cls.from_pandas(pandas.read_clipboard(sep=sep, **kwargs))

    @classmethod
    @_inherit_docstrings(pandas.read_excel, apilink="pandas.read_excel")
    @doc(
        _doc_default_io_method,
        summary="Read an Excel file into query compiler",
        returns="""BaseQueryCompiler or dict :
    QueryCompiler or dict with read data.""",
    )
    def read_excel(cls, **kwargs):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_excel`")
        if isinstance(kwargs["io"], ExcelFile):
            # otherwise, Modin objects may be passed to the pandas context, resulting
            # in undefined behavior
            # for example in the case: pd.read_excel(pd.ExcelFile), since reading from
            # pd.ExcelFile in `read_excel` isn't supported
            kwargs["io"]._set_pandas_mode()
        intermediate = pandas.read_excel(**kwargs)
        if isinstance(intermediate, dict):
            parsed = type(intermediate)()
            for key in intermediate.keys():
                parsed[key] = cls.from_pandas(intermediate.get(key))
            return parsed
        else:
            return cls.from_pandas(intermediate)

    @classmethod
    @_inherit_docstrings(pandas.read_hdf, apilink="pandas.read_hdf")
    @doc(
        _doc_default_io_method,
        summary="Read data from hdf store into query compiler",
        returns=_doc_returns_qc,
    )
    def read_hdf(
        cls,
        path_or_buf,
        key=None,
        mode: str = "r",
        errors: str = "strict",
        where=None,
        start=None,
        stop=None,
        columns=None,
        iterator=False,
        chunksize=None,
        **kwargs,
    ):  # noqa: PR01
        from modin.pandas.io import HDFStore

        ErrorMessage.default_to_pandas("`read_hdf`")
        modin_store = isinstance(path_or_buf, HDFStore)
        if modin_store:
            path_or_buf._return_modin_dataframe = False
        df = pandas.read_hdf(
            path_or_buf,
            key=key,
            mode=mode,
            columns=columns,
            errors=errors,
            where=where,
            start=start,
            stop=stop,
            iterator=iterator,
            chunksize=chunksize,
            **kwargs,
        )
        if modin_store:
            path_or_buf._return_modin_dataframe = True

        return cls.from_pandas(df)

    @classmethod
    @_inherit_docstrings(pandas.read_feather, apilink="pandas.read_feather")
    @doc(
        _doc_default_io_method,
        summary="Load a feather-format object from the file path into query compiler",
        returns=_doc_returns_qc,
    )
    def read_feather(
        cls,
        path,
        **kwargs,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_feather`")
        return cls.from_pandas(
            pandas.read_feather(
                path,
                **kwargs,
            )
        )

    @classmethod
    @_inherit_docstrings(pandas.read_stata, apilink="pandas.read_stata")
    @doc(
        _doc_default_io_method,
        summary="Read Stata file into query compiler",
        returns=_doc_returns_qc,
    )
    def read_stata(
        cls,
        filepath_or_buffer,
        **kwargs,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_stata`")
        return cls.from_pandas(pandas.read_stata(filepath_or_buffer, **kwargs))

    @classmethod
    @_inherit_docstrings(pandas.read_sas, apilink="pandas.read_sas")
    @doc(
        _doc_default_io_method,
        summary="Read SAS files stored as either XPORT or SAS7BDAT format files\ninto query compiler",
        returns=_doc_returns_qc,
    )
    def read_sas(
        cls,
        filepath_or_buffer,
        *,
        format=None,
        index=None,
        encoding=None,
        chunksize=None,
        iterator=False,
        **kwargs,
    ):  # pragma: no cover # noqa: PR01
        ErrorMessage.default_to_pandas("`read_sas`")
        return cls.from_pandas(
            pandas.read_sas(
                filepath_or_buffer,
                format=format,
                index=index,
                encoding=encoding,
                chunksize=chunksize,
                iterator=iterator,
                **kwargs,
            )
        )

    @classmethod
    @_inherit_docstrings(pandas.read_pickle, apilink="pandas.read_pickle")
    @doc(
        _doc_default_io_method,
        summary="Load pickled pandas object (or any object) from file into query compiler",
        returns=_doc_returns_qc,
    )
    def read_pickle(
        cls,
        filepath_or_buffer,
        **kwargs,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_pickle`")
        return cls.from_pandas(
            pandas.read_pickle(
                filepath_or_buffer,
                **kwargs,
            )
        )

    @classmethod
    @_inherit_docstrings(pandas.read_sql, apilink="pandas.read_sql")
    @doc(
        _doc_default_io_method,
        summary="Read SQL query or database table into query compiler",
        returns=_doc_returns_qc,
    )
    def read_sql(
        cls,
        sql,
        con,
        index_col=None,
        coerce_float=True,
        params=None,
        parse_dates=None,
        columns=None,
        chunksize=None,
        dtype_backend=no_default,
        dtype=None,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_sql`")
        if isinstance(con, ModinDatabaseConnection):
            con = con.get_connection()
        result = pandas.read_sql(
            sql,
            con,
            index_col=index_col,
            coerce_float=coerce_float,
            params=params,
            parse_dates=parse_dates,
            columns=columns,
            chunksize=chunksize,
            dtype_backend=dtype_backend,
            dtype=dtype,
        )

        if isinstance(result, (pandas.DataFrame, pandas.Series)):
            return cls.from_pandas(result)
        return (cls.from_pandas(df) for df in result)

    @classmethod
    @_inherit_docstrings(pandas.read_fwf, apilink="pandas.read_fwf")
    @doc(
        _doc_default_io_method,
        summary="Read a table of fixed-width formatted lines into query compiler",
        returns=_doc_returns_qc_or_parser,
    )
    def read_fwf(
        cls,
        filepath_or_buffer,
        *,
        colspecs="infer",
        widths=None,
        infer_nrows=100,
        dtype_backend=no_default,
        iterator=False,
        chunksize=None,
        **kwds,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_fwf`")
        pd_obj = pandas.read_fwf(
            filepath_or_buffer,
            colspecs=colspecs,
            widths=widths,
            infer_nrows=infer_nrows,
            dtype_backend=dtype_backend,
            iterator=iterator,
            chunksize=chunksize,
            **kwds,
        )
        if isinstance(pd_obj, pandas.DataFrame):
            return cls.from_pandas(pd_obj)
        if isinstance(pd_obj, pandas.io.parsers.TextFileReader):
            # Overwriting the read method should return a Modin DataFrame for calls
            # to __next__ and get_chunk
            pd_read = pd_obj.read
            pd_obj.read = lambda *args, **kwargs: cls.from_pandas(
                pd_read(*args, **kwargs)
            )
        return pd_obj

    @classmethod
    @_inherit_docstrings(pandas.read_sql_table, apilink="pandas.read_sql_table")
    @doc(
        _doc_default_io_method,
        summary="Read SQL database table into query compiler",
        returns=_doc_returns_qc,
    )
    def read_sql_table(
        cls,
        table_name,
        con,
        schema=None,
        index_col=None,
        coerce_float=True,
        parse_dates=None,
        columns=None,
        chunksize=None,
        dtype_backend=no_default,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_sql_table`")
        return cls.from_pandas(
            pandas.read_sql_table(
                table_name,
                con,
                schema=schema,
                index_col=index_col,
                coerce_float=coerce_float,
                parse_dates=parse_dates,
                columns=columns,
                chunksize=chunksize,
                dtype_backend=dtype_backend,
            )
        )

    @classmethod
    @_inherit_docstrings(pandas.read_sql_query, apilink="pandas.read_sql_query")
    @doc(
        _doc_default_io_method,
        summary="Read SQL query into query compiler",
        returns=_doc_returns_qc,
    )
    def read_sql_query(
        cls,
        sql,
        con,
        **kwargs,
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_sql_query`")
        return cls.from_pandas(
            pandas.read_sql_query(
                sql,
                con,
                **kwargs,
            )
        )

    @classmethod
    @_inherit_docstrings(pandas.read_spss, apilink="pandas.read_spss")
    @doc(
        _doc_default_io_method,
        summary="Load an SPSS file from the file path, returning a query compiler",
        returns=_doc_returns_qc,
    )
    def read_spss(
        cls, path, usecols, convert_categoricals, dtype_backend
    ):  # noqa: PR01
        ErrorMessage.default_to_pandas("`read_spss`")
        return cls.from_pandas(
            pandas.read_spss(
                path,
                usecols=usecols,
                convert_categoricals=convert_categoricals,
                dtype_backend=dtype_backend,
            )
        )

    @classmethod
    @_inherit_docstrings(pandas.DataFrame.to_sql, apilink="pandas.DataFrame.to_sql")
    def to_sql(
        cls,
        qc,
        name,
        con,
        schema=None,
        if_exists="fail",
        index=True,
        index_label=None,
        chunksize=None,
        dtype=None,
        method=None,
    ):  # noqa: PR01
        """
        Write records stored in a DataFrame to a SQL database using pandas.

        For parameters description please refer to pandas API.
        """
        ErrorMessage.default_to_pandas("`to_sql`")
        df = qc.to_pandas()
        df.to_sql(
            name=name,
            con=con,
            schema=schema,
            if_exists=if_exists,
            index=index,
            index_label=index_label,
            chunksize=chunksize,
            dtype=dtype,
            method=method,
        )

    @classmethod
    @_inherit_docstrings(
        pandas.DataFrame.to_pickle, apilink="pandas.DataFrame.to_pickle"
    )
    def to_pickle(
        cls,
        obj: Any,
        filepath_or_buffer,
        **kwargs,
    ):  # noqa: PR01, D200
        """
        Pickle (serialize) object to file.
        """
        ErrorMessage.default_to_pandas("`to_pickle`")
        if isinstance(obj, BaseQueryCompiler):
            obj = obj.to_pandas()

        return pandas.to_pickle(
            obj,
            filepath_or_buffer=filepath_or_buffer,
            **kwargs,
        )

    @classmethod
    @_inherit_docstrings(pandas.DataFrame.to_csv, apilink="pandas.DataFrame.to_csv")
    def to_csv(cls, obj, **kwargs):  # noqa: PR01
        """
        Write object to a comma-separated values (CSV) file using pandas.

        For parameters description please refer to pandas API.
        """
        ErrorMessage.default_to_pandas("`to_csv`")
        if isinstance(obj, BaseQueryCompiler):
            obj = obj.to_pandas()

        return obj.to_csv(**kwargs)

    @classmethod
    @_inherit_docstrings(pandas.DataFrame.to_json, apilink="pandas.DataFrame.to_json")
    def to_json(cls, obj, path, **kwargs):  # noqa: PR01
        """
        Convert the object to a JSON string.

        For parameters description please refer to pandas API.
        """
        ErrorMessage.default_to_pandas("`to_json`")
        if isinstance(obj, BaseQueryCompiler):
            obj = obj.to_pandas()

        return obj.to_json(path, **kwargs)

    @classmethod
    @_inherit_docstrings(pandas.DataFrame.to_xml, apilink="pandas.DataFrame.to_xml")
    def to_xml(cls, obj, path_or_buffer, **kwargs):  # noqa: PR01
        """
        Convert the object to a XML string.

        For parameters description please refer to pandas API.
        """
        ErrorMessage.default_to_pandas("`to_xml`")
        if isinstance(obj, BaseQueryCompiler):
            obj = obj.to_pandas()

        return obj.to_xml(path_or_buffer, **kwargs)

    @classmethod
    @_inherit_docstrings(
        pandas.DataFrame.to_parquet, apilink="pandas.DataFrame.to_parquet"
    )
    def to_parquet(cls, obj, path, **kwargs):  # noqa: PR01
        """
        Write object to the binary parquet format using pandas.

        For parameters description please refer to pandas API.
        """
        ErrorMessage.default_to_pandas("`to_parquet`")
        if isinstance(obj, BaseQueryCompiler):
            obj = obj.to_pandas()

        return obj.to_parquet(path, **kwargs)

    @classmethod
    def to_ray_dataset(cls, modin_obj):
        """
        Convert a Modin DataFrame/Series to a Ray Dataset.

        Parameters
        ----------
        modin_obj : modin.pandas.DataFrame, modin.pandas.Series
            The Modin DataFrame/Series to convert.

        Returns
        -------
        ray.data.Dataset
            Converted object with type depending on input.

        Notes
        -----
        Modin DataFrame/Series can only be converted to a Ray Dataset if Modin uses a Ray engine.
        If another engine is used, the runtime exception will be raised.
        """
        raise RuntimeError(
            "Modin Dataframe can only be converted to a Ray Dataset if Modin uses a Ray engine."
        )



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""Module contains class ``ModinDtypes``."""

from typing import TYPE_CHECKING, Callable, Optional, Union

import numpy as np
import pandas
from pandas._typing import IndexLabel
from pandas.core.dtypes.cast import find_common_type

if TYPE_CHECKING:
    from modin.core.dataframe.pandas.dataframe.dataframe import PandasDataframe
    from .index import ModinIndex

from modin.error_message import ErrorMessage


class DtypesDescriptor:
    """
    Describes partial dtypes.

    Parameters
    ----------
    known_dtypes : dict[IndexLabel, np.dtype] or pandas.Series, optional
        Columns that we know dtypes for.
    cols_with_unknown_dtypes : list[IndexLabel], optional
        Column names that have unknown dtypes. If specified together with `remaining_dtype`, must describe all
        columns with unknown dtypes, otherwise, the missing columns will be assigned to `remaining_dtype`.
        If `cols_with_unknown_dtypes` is incomplete, you must specify `know_all_names=False`.
    remaining_dtype : np.dtype, optional
        Dtype for columns that are not present neither in `known_dtypes` nor in `cols_with_unknown_dtypes`.
        This parameter is intended to describe columns that we known dtypes for, but don't know their
        names yet. Note, that this parameter DOESN'T describe dtypes for columns from `cols_with_unknown_dtypes`.
    parent_df : PandasDataframe, optional
        Dataframe object for which we describe dtypes. This dataframe will be used to compute
        missing dtypes on ``.materialize()``.
    columns_order : dict[int, IndexLabel], optional
        Order of columns in the dataframe. If specified, must describe all the columns of the dataframe.
    know_all_names : bool, default: True
        Whether `known_dtypes` and `cols_with_unknown_dtypes` contain all column names for this dataframe besides those,
        that are being described by `remaining_dtype`.
        One can't pass `know_all_names=False` together with `remaining_dtype` as this creates ambiguity
        on how to interpret missing columns (whether they belong to `remaining_dtype` or not).
    _schema_is_known : bool, optional
        Whether `known_dtypes` describe all columns in the dataframe. This parameter intended mostly
        for internal use.
    """

    def __init__(
        self,
        known_dtypes: Optional[Union[dict[IndexLabel, np.dtype], pandas.Series]] = None,
        cols_with_unknown_dtypes: Optional[list[IndexLabel]] = None,
        remaining_dtype: Optional[np.dtype] = None,
        parent_df: Optional["PandasDataframe"] = None,
        columns_order: Optional[dict[int, IndexLabel]] = None,
        know_all_names: bool = True,
        _schema_is_known: Optional[bool] = None,
    ):
        if not know_all_names and remaining_dtype is not None:
            raise ValueError(
                "It's not allowed to pass 'remaining_dtype' and 'know_all_names=False' at the same time."
            )
        # columns with known dtypes
        self._known_dtypes: dict[IndexLabel, np.dtype] = (
            {} if known_dtypes is None else dict(known_dtypes)
        )
        if known_dtypes is not None and len(self._known_dtypes) != len(known_dtypes):
            raise NotImplementedError(
                "Duplicated column names are not yet supported by DtypesDescriptor"
            )
        # columns with unknown dtypes (they're not described by 'remaining_dtype')
        if cols_with_unknown_dtypes is not None and len(
            set(cols_with_unknown_dtypes)
        ) != len(cols_with_unknown_dtypes):
            raise NotImplementedError(
                "Duplicated column names are not yet supported by DtypesDescriptor"
            )
        self._cols_with_unknown_dtypes: list[IndexLabel] = (
            [] if cols_with_unknown_dtypes is None else cols_with_unknown_dtypes
        )
        # whether 'known_dtypes' describe all columns in the dataframe
        self._schema_is_known: Optional[bool] = _schema_is_known
        if self._schema_is_known is None:
            self._schema_is_known = False
            if (
                # if 'cols_with_unknown_dtypes' was explicitly specified as an empty list and
                # we don't have any 'remaining_dtype', then we assume that 'known_dtypes' are complete
                cols_with_unknown_dtypes is not None
                and know_all_names
                and remaining_dtype is None
                and len(self._known_dtypes) > 0
            ):
                self._schema_is_known = len(cols_with_unknown_dtypes) == 0

        self._know_all_names: bool = know_all_names
        # a common dtype for columns that are not present in 'known_dtypes' nor in 'cols_with_unknown_dtypes'
        self._remaining_dtype: Optional[np.dtype] = remaining_dtype
        self._parent_df: Optional["PandasDataframe"] = parent_df
        if columns_order is None:
            self._columns_order: Optional[dict[int, IndexLabel]] = None
            # try to compute '._columns_order' using 'parent_df'
            self.columns_order
        else:
            if remaining_dtype is not None:
                raise ValueError(
                    "Passing 'columns_order' and 'remaining_dtype' is ambiguous. You have to manually "
                    + "complete 'known_dtypes' using the information from 'columns_order' and 'remaining_dtype'."
                )
            elif not self._know_all_names:
                raise ValueError(
                    "Passing 'columns_order' and 'know_all_names=False' is ambiguous. You have to manually "
                    + "complete 'cols_with_unknown_dtypes' using the information from 'columns_order' "
                    + "and pass 'know_all_names=True'."
                )
            elif len(columns_order) != (
                len(self._cols_with_unknown_dtypes) + len(self._known_dtypes)
            ):
                raise ValueError(
                    "The length of 'columns_order' doesn't match to 'known_dtypes' and 'cols_with_unknown_dtypes'"
                )
            self._columns_order: Optional[dict[int, IndexLabel]] = columns_order

    def update_parent(self, new_parent: "PandasDataframe"):
        """
        Set new parent dataframe.

        Parameters
        ----------
        new_parent : PandasDataframe
        """
        self._parent_df = new_parent
        LazyProxyCategoricalDtype.update_dtypes(self._known_dtypes, new_parent)
        # try to compute '._columns_order' using 'new_parent'
        self.columns_order

    @property
    def columns_order(self) -> Optional[dict[int, IndexLabel]]:
        """
        Get order of columns for the described dataframe if available.

        Returns
        -------
        dict[int, IndexLabel] or None
        """
        if self._columns_order is not None:
            return self._columns_order
        if self._parent_df is None or not self._parent_df.has_materialized_columns:
            return None

        actual_columns = self._parent_df.columns
        self._normalize_self_levels(actual_columns)

        self._columns_order = {i: col for i, col in enumerate(actual_columns)}
        # we got information about new columns and thus can potentially
        # extend our knowledge about missing dtypes
        if len(self._columns_order) > (
            len(self._known_dtypes) + len(self._cols_with_unknown_dtypes)
        ):
            new_cols = [
                col
                for col in self._columns_order.values()
                if col not in self._known_dtypes
                and col not in self._cols_with_unknown_dtypes
            ]
            if self._remaining_dtype is not None:
                self._known_dtypes.update(
                    {col: self._remaining_dtype for col in new_cols}
                )
                self._remaining_dtype = None
                if len(self._cols_with_unknown_dtypes) == 0:
                    self._schema_is_known = True
            else:
                self._cols_with_unknown_dtypes.extend(new_cols)
        self._know_all_names = True
        return self._columns_order

    def __repr__(self):  # noqa: GL08
        return (
            f"DtypesDescriptor:\n\tknown dtypes: {self._known_dtypes};\n\t"
            + f"remaining dtype: {self._remaining_dtype};\n\t"
            + f"cols with unknown dtypes: {self._cols_with_unknown_dtypes};\n\t"
            + f"schema is known: {self._schema_is_known};\n\t"
            + f"has parent df: {self._parent_df is not None};\n\t"
            + f"columns order: {self._columns_order};\n\t"
            + f"know all names: {self._know_all_names}"
        )

    def __str__(self):  # noqa: GL08
        return self.__repr__()

    def lazy_get(
        self, ids: list[Union[IndexLabel, int]], numeric_index: bool = False
    ) -> "DtypesDescriptor":
        """
        Get dtypes descriptor for a subset of columns without triggering any computations.

        Parameters
        ----------
        ids : list of index labels or positional indexers
            Columns for the subset.
        numeric_index : bool, default: False
            Whether `ids` are positional indixes or column labels to take.

        Returns
        -------
        DtypesDescriptor
            Descriptor that describes dtypes for columns specified in `ids`.
        """
        if len(set(ids)) != len(ids):
            raise NotImplementedError(
                "Duplicated column names are not yet supported by DtypesDescriptor"
            )

        if numeric_index:
            if self.columns_order is not None:
                ids = [self.columns_order[i] for i in ids]
            else:
                raise ValueError(
                    "Can't lazily get columns by positional indixers if the columns order is unknown"
                )

        result = {}
        unknown_cols = []
        columns_order = {}
        for i, col in enumerate(ids):
            columns_order[i] = col
            if col in self._cols_with_unknown_dtypes:
                unknown_cols.append(col)
                continue
            dtype = self._known_dtypes.get(col)
            if dtype is None and self._remaining_dtype is None:
                unknown_cols.append(col)
            elif dtype is None and self._remaining_dtype is not None:
                result[col] = self._remaining_dtype
            else:
                result[col] = dtype
        remaining_dtype = self._remaining_dtype if len(unknown_cols) != 0 else None
        return DtypesDescriptor(
            result,
            unknown_cols,
            remaining_dtype,
            self._parent_df,
            columns_order=columns_order,
        )

    def copy(self) -> "DtypesDescriptor":
        """
        Get a copy of this descriptor.

        Returns
        -------
        DtypesDescriptor
        """
        return type(self)(
            # should access '.columns_order' first, as it may compute columns order
            # and complete the metadata for 'self'
            columns_order=(
                None if self.columns_order is None else self.columns_order.copy()
            ),
            known_dtypes=self._known_dtypes.copy(),
            cols_with_unknown_dtypes=self._cols_with_unknown_dtypes.copy(),
            remaining_dtype=self._remaining_dtype,
            parent_df=self._parent_df,
            know_all_names=self._know_all_names,
            _schema_is_known=self._schema_is_known,
        )

    def set_index(
        self, new_index: Union[pandas.Index, "ModinIndex"]
    ) -> "DtypesDescriptor":
        """
        Set new column names for this descriptor.

        Parameters
        ----------
        new_index : pandas.Index or ModinIndex

        Returns
        -------
        DtypesDescriptor
            New descriptor with updated column names.

        Notes
        -----
        Calling this method on a descriptor that returns ``None`` for ``.columns_order``
        will result into information lose.
        """
        if len(new_index) != len(set(new_index)):
            raise NotImplementedError(
                "Duplicated column names are not yet supported by DtypesDescriptor"
            )

        if self.columns_order is None:
            # we can't map new columns to old columns and lost all dtypes :(
            return DtypesDescriptor(
                cols_with_unknown_dtypes=new_index,
                columns_order={i: col for i, col in enumerate(new_index)},
                parent_df=self._parent_df,
                know_all_names=True,
            )

        new_self = self.copy()
        renamer = {old_c: new_index[i] for i, old_c in new_self.columns_order.items()}
        new_self._known_dtypes = {
            renamer[old_col]: value for old_col, value in new_self._known_dtypes.items()
        }
        new_self._cols_with_unknown_dtypes = [
            renamer[old_col] for old_col in new_self._cols_with_unknown_dtypes
        ]
        new_self._columns_order = {
            i: renamer[old_col] for i, old_col in new_self._columns_order.items()
        }
        return new_self

    def equals(self, other: "DtypesDescriptor") -> bool:
        """
        Compare two descriptors for equality.

        Parameters
        ----------
        other : DtypesDescriptor

        Returns
        -------
        bool
        """
        return (
            self._known_dtypes == other._known_dtypes
            and set(self._cols_with_unknown_dtypes)
            == set(other._cols_with_unknown_dtypes)
            and self._remaining_dtype == other._remaining_dtype
            and self._schema_is_known == other._schema_is_known
            and self.columns_order == other.columns_order
            and self._know_all_names == other._know_all_names
        )

    @property
    def is_materialized(self) -> bool:
        """
        Whether this descriptor contains information about all dtypes in the dataframe.

        Returns
        -------
        bool
        """
        return self._schema_is_known

    def _materialize_all_names(self):
        """Materialize missing column names."""
        if self._know_all_names:
            return

        all_cols = self._parent_df.columns
        self._normalize_self_levels(all_cols)
        for col in all_cols:
            if (
                col not in self._known_dtypes
                and col not in self._cols_with_unknown_dtypes
            ):
                self._cols_with_unknown_dtypes.append(col)

        self._know_all_names = True

    def _materialize_cols_with_unknown_dtypes(self):
        """Compute dtypes for cols specified in `._cols_with_unknown_dtypes`."""
        if (
            len(self._known_dtypes) == 0
            and len(self._cols_with_unknown_dtypes) == 0
            and not self._know_all_names
        ):
            # here we have to compute dtypes for all columns in the dataframe,
            # so avoiding columns materialization by setting 'subset=None'
            subset = None
        else:
            if not self._know_all_names:
                self._materialize_all_names()
            subset = self._cols_with_unknown_dtypes

        if subset is None or len(subset) > 0:
            self._known_dtypes.update(self._parent_df._compute_dtypes(subset))

        self._know_all_names = True
        self._cols_with_unknown_dtypes = []

    def materialize(self):
        """Complete information about dtypes."""
        if self.is_materialized:
            return
        if self._parent_df is None:
            raise RuntimeError(
                "It's not allowed to call '.materialize()' before '._parent_df' is specified."
            )

        self._materialize_cols_with_unknown_dtypes()

        if self._remaining_dtype is not None:
            cols = self._parent_df.columns
            self._normalize_self_levels(cols)
            self._known_dtypes.update(
                {
                    col: self._remaining_dtype
                    for col in cols
                    if col not in self._known_dtypes
                }
            )

        # we currently not guarantee for dtypes to be in a proper order:
        # https://github.com/modin-project/modin/blob/8a332c1597c54d36f7ccbbd544e186b689f9ceb1/modin/pandas/test/utils.py#L644-L646
        # so restoring the order only if it's possible
        if self.columns_order is not None:
            assert len(self.columns_order) == len(self._known_dtypes)
            self._known_dtypes = {
                self.columns_order[i]: self._known_dtypes[self.columns_order[i]]
                for i in range(len(self.columns_order))
            }

        self._schema_is_known = True
        self._remaining_dtype = None
        self._parent_df = None

    def to_series(self) -> pandas.Series:
        """
        Convert descriptor to a pandas Series.

        Returns
        -------
        pandas.Series
        """
        self.materialize()
        return pandas.Series(self._known_dtypes)

    def get_dtypes_set(self) -> set[np.dtype]:
        """
        Get a set of dtypes from the descriptor.

        Returns
        -------
        set[np.dtype]
        """
        if len(self._cols_with_unknown_dtypes) > 0 or not self._know_all_names:
            self._materialize_cols_with_unknown_dtypes()
        known_dtypes: set[np.dtype] = set(self._known_dtypes.values())
        if self._remaining_dtype is not None:
            known_dtypes.add(self._remaining_dtype)
        return known_dtypes

    @classmethod
    def _merge_dtypes(
        cls, values: list[Union["DtypesDescriptor", pandas.Series, None]]
    ) -> "DtypesDescriptor":
        """
        Union columns described by ``values`` and compute common dtypes for them.

        Parameters
        ----------
        values : list of DtypesDescriptors, pandas.Series or Nones

        Returns
        -------
        DtypesDescriptor
        """
        known_dtypes = {}
        cols_with_unknown_dtypes = []
        know_all_names = True
        dtypes_are_unknown = False

        # index - joined column names, columns - dtypes taken from 'values'
        #        0     1     2      3
        #  col1  int   bool  float  int
        #  col2  int   int   int    int
        #  colN  bool  bool  bool   int
        dtypes_matrix = pandas.DataFrame()

        for i, val in enumerate(values):
            if isinstance(val, cls):
                know_all_names &= val._know_all_names
                dtypes = val._known_dtypes.copy()
                dtypes.update({col: "unknown" for col in val._cols_with_unknown_dtypes})
                if val._remaining_dtype is not None:
                    # we can't process remaining dtypes, so just discarding them
                    know_all_names = False

                # setting a custom name to the Series to prevent duplicated names
                # in the 'dtypes_matrix'
                series = pandas.Series(dtypes, name=i)
                dtypes_matrix = pandas.concat([dtypes_matrix, series], axis=1)
                dtypes_matrix.fillna(
                    value={
                        # If we encountered a 'NaN' while 'val' describes all the columns, then
                        # it means, that the missing columns for this instance will be filled with NaNs (floats),
                        # otherwise, it may indicate missing columns that this 'val' has no info about,
                        # meaning that we shouldn't try computing a new dtype for this column,
                        # so marking it as 'unknown'
                        i: (
                            np.dtype(float)
                            if val._know_all_names and val._remaining_dtype is None
                            else "unknown"
                        )
                    },
                    inplace=True,
                )
            elif isinstance(val, pandas.Series):
                dtypes_matrix = pandas.concat([dtypes_matrix, val], axis=1)
            elif val is None:
                # one of the 'dtypes' is None, meaning that we wouldn't been infer a valid result dtype,
                # however, we're continuing our loop so we would at least know the columns we're missing
                # dtypes for
                dtypes_are_unknown = True
                know_all_names = False
            else:
                raise NotImplementedError(type(val))

        if dtypes_are_unknown:
            return DtypesDescriptor(
                cols_with_unknown_dtypes=dtypes_matrix.index.tolist(),
                know_all_names=know_all_names,
            )

        def combine_dtypes(row):
            if (row == "unknown").any():
                return "unknown"
            row = row.fillna(np.dtype("float"))
            return find_common_type(list(row.values))

        dtypes = dtypes_matrix.apply(combine_dtypes, axis=1)

        for col, dtype in dtypes.items():
            if dtype == "unknown":
                cols_with_unknown_dtypes.append(col)
            else:
                known_dtypes[col] = dtype

        return DtypesDescriptor(
            known_dtypes,
            cols_with_unknown_dtypes,
            remaining_dtype=None,
            know_all_names=know_all_names,
        )

    @classmethod
    def concat(
        cls, values: list[Union["DtypesDescriptor", pandas.Series, None]], axis: int = 0
    ) -> "DtypesDescriptor":
        """
        Concatenate dtypes descriptors into a single descriptor.

        Parameters
        ----------
        values : list of DtypesDescriptors and pandas.Series
        axis : int, default: 0
            If ``axis == 0``: concatenate column names. This implements the logic of
            how dtypes are combined on ``pd.concat([df1, df2], axis=1)``.
            If ``axis == 1``: perform a union join for the column names described by
            `values` and then find common dtypes for the columns appeared to be in
            an intersection. This implements the logic of how dtypes are combined on
            ``pd.concat([df1, df2], axis=0).dtypes``.

        Returns
        -------
        DtypesDescriptor
        """
        if axis == 1:
            return cls._merge_dtypes(values)
        known_dtypes = {}
        cols_with_unknown_dtypes = []
        schema_is_known = True
        # some default value to not mix it with 'None'
        remaining_dtype = "default"
        know_all_names = True

        for val in values:
            if isinstance(val, cls):
                all_new_cols = (
                    list(val._known_dtypes.keys()) + val._cols_with_unknown_dtypes
                )
                if any(
                    col in known_dtypes or col in cols_with_unknown_dtypes
                    for col in all_new_cols
                ):
                    raise NotImplementedError(
                        "Duplicated column names are not yet supported by DtypesDescriptor"
                    )
                know_all_names &= val._know_all_names
                known_dtypes.update(val._known_dtypes)
                cols_with_unknown_dtypes.extend(val._cols_with_unknown_dtypes)
                if know_all_names:
                    if (
                        remaining_dtype == "default"
                        and val._remaining_dtype is not None
                    ):
                        remaining_dtype = val._remaining_dtype
                    elif (
                        remaining_dtype != "default"
                        and val._remaining_dtype is not None
                        and remaining_dtype != val._remaining_dtype
                    ):
                        remaining_dtype = None
                        know_all_names = False
                else:
                    remaining_dtype = None
                schema_is_known &= val._schema_is_known
            elif isinstance(val, pandas.Series):
                if any(
                    col in known_dtypes or col in cols_with_unknown_dtypes
                    for col in val.index
                ):
                    raise NotImplementedError(
                        "Duplicated column names are not yet supported by DtypesDescriptor"
                    )
                known_dtypes.update(val)
            elif val is None:
                remaining_dtype = None
                schema_is_known = False
                know_all_names = False
            else:
                raise NotImplementedError(type(val))
        return cls(
            known_dtypes,
            cols_with_unknown_dtypes,
            None if remaining_dtype == "default" else remaining_dtype,
            parent_df=None,
            _schema_is_known=schema_is_known,
            know_all_names=know_all_names,
        )

    @staticmethod
    def _normalize_levels(columns, reference=None):
        """
        Normalize levels of MultiIndex column names.

        The function fills missing levels with empty strings as pandas do:
        '''
        >>> columns = ["a", ("l1", "l2"), ("l1a", "l2a", "l3a")]
        >>> _normalize_levels(columns)
        [("a", "", ""), ("l1", "l2", ""), ("l1a", "l2a", "l3a")]
        >>> # with a reference
        >>> idx = pandas.MultiIndex(...)
        >>> idx.nlevels
        4
        >>> _normalize_levels(columns, reference=idx)
        [("a", "", "", ""), ("l1", "l2", "", ""), ("l1a", "l2a", "l3a", "")]
        '''

        Parameters
        ----------
        columns : sequence
            Labels to normalize. If dictionary, will replace keys with normalized columns.
        reference : pandas.Index, optional
            An index to match the number of levels with. If reference is a MultiIndex, then the reference number
            of levels should not be greater than the maximum number of levels in `columns`. If not specified,
            the `columns` themselves become a `reference`.

        Returns
        -------
        sequence
            Column values with normalized levels.
        dict[hashable, hashable]
            Mapping from old column names to new names, only contains column names that
            were changed.

        Raises
        ------
        ValueError
            When the reference number of levels is greater than the maximum number of levels
            in `columns`.
        """
        if reference is None:
            reference = columns

        if isinstance(reference, pandas.Index):
            max_nlevels = reference.nlevels
        else:
            max_nlevels = 1
            for col in reference:
                if isinstance(col, tuple):
                    max_nlevels = max(max_nlevels, len(col))

        # if the reference is a regular flat index, then no actions are required (the result will be
        # a flat index containing tuples of different lengths, this behavior fully matches pandas).
        # Yes, this shortcut skips the 'if max_columns_nlevels > max_nlevels' below check on purpose.
        if max_nlevels == 1:
            return columns, {}

        max_columns_nlevels = 1
        for col in columns:
            if isinstance(col, tuple):
                max_columns_nlevels = max(max_columns_nlevels, len(col))

        if max_columns_nlevels > max_nlevels:
            raise ValueError(
                f"The reference number of levels is greater than the maximum number of levels in columns: {max_columns_nlevels} > {max_nlevels}"
            )

        new_columns = []
        old_to_new_mapping = {}
        for col in columns:
            old_col = col
            if not isinstance(col, tuple):
                col = (col,)
            col = col + ("",) * (max_nlevels - len(col))
            new_columns.append(col)
            if old_col != col:
                old_to_new_mapping[old_col] = col

        return new_columns, old_to_new_mapping

    def _normalize_self_levels(self, reference=None):
        """
        Call ``self._normalize_levels()`` for known and unknown dtypes of this object.

        Parameters
        ----------
        reference : pandas.Index, optional
        """
        _, old_to_new_mapping = self._normalize_levels(
            self._known_dtypes.keys(), reference
        )
        for old_col, new_col in old_to_new_mapping.items():
            value = self._known_dtypes.pop(old_col)
            self._known_dtypes[new_col] = value
        self._cols_with_unknown_dtypes, _ = self._normalize_levels(
            self._cols_with_unknown_dtypes, reference
        )


class ModinDtypes:
    """
    A class that hides the various implementations of the dtypes needed for optimization.

    Parameters
    ----------
    value : pandas.Series, callable, DtypesDescriptor or ModinDtypes, optional
    """

    def __init__(
        self,
        value: Optional[
            Union[Callable, pandas.Series, DtypesDescriptor, "ModinDtypes"]
        ],
    ):
        if callable(value) or isinstance(value, pandas.Series):
            self._value = value
        elif isinstance(value, DtypesDescriptor):
            self._value = value.to_series() if value.is_materialized else value
        elif isinstance(value, type(self)):
            self._value = value.copy()._value
        elif isinstance(value, None):
            self._value = DtypesDescriptor()
        else:
            raise ValueError(f"ModinDtypes doesn't work with '{value}'")

    def __repr__(self):  # noqa: GL08
        return f"ModinDtypes:\n\tvalue type: {type(self._value)};\n\tvalue:\n\t{self._value}"

    def __str__(self):  # noqa: GL08
        return self.__repr__()

    @property
    def is_materialized(self) -> bool:
        """
        Check if the internal representation is materialized.

        Returns
        -------
        bool
        """
        return isinstance(self._value, pandas.Series)

    def get_dtypes_set(self) -> set[np.dtype]:
        """
        Get a set of dtypes from the descriptor.

        Returns
        -------
        set[np.dtype]
        """
        if isinstance(self._value, DtypesDescriptor):
            return self._value.get_dtypes_set()
        if not self.is_materialized:
            self.get()
        return set(self._value.values)

    def maybe_specify_new_frame_ref(
        self, new_parent: "PandasDataframe"
    ) -> "ModinDtypes":
        """
        Set a new parent for the stored value if needed.

        Parameters
        ----------
        new_parent : PandasDataframe

        Returns
        -------
        ModinDtypes
            A copy of ``ModinDtypes`` with updated parent.
        """
        new_self = self.copy()
        if new_self.is_materialized:
            LazyProxyCategoricalDtype.update_dtypes(new_self._value, new_parent)
            return new_self
        if isinstance(self._value, DtypesDescriptor):
            new_self._value.update_parent(new_parent)
            return new_self
        return new_self

    def lazy_get(self, ids: list, numeric_index: bool = False) -> "ModinDtypes":
        """
        Get new ``ModinDtypes`` for a subset of columns without triggering any computations.

        Parameters
        ----------
        ids : list of index labels or positional indexers
            Columns for the subset.
        numeric_index : bool, default: False
            Whether `ids` are positional indixes or column labels to take.

        Returns
        -------
        ModinDtypes
            ``ModinDtypes`` that describes dtypes for columns specified in `ids`.
        """
        if isinstance(self._value, DtypesDescriptor):
            res = self._value.lazy_get(ids, numeric_index)
            return ModinDtypes(res)
        elif callable(self._value):
            new_self = self.copy()
            old_value = new_self._value
            new_self._value = lambda: (
                old_value().iloc[ids] if numeric_index else old_value()[ids]
            )
            return new_self
        ErrorMessage.catch_bugs_and_request_email(
            failure_condition=not self.is_materialized
        )
        return ModinDtypes(self._value.iloc[ids] if numeric_index else self._value[ids])

    @classmethod
    def concat(cls, values: list, axis: int = 0) -> "ModinDtypes":
        """
        Concatenate dtypes.

        Parameters
        ----------
        values : list of DtypesDescriptors, pandas.Series, ModinDtypes and Nones
        axis : int, default: 0
            If ``axis == 0``: concatenate column names. This implements the logic of
            how dtypes are combined on ``pd.concat([df1, df2], axis=1)``.
            If ``axis == 1``: perform a union join for the column names described by
            `values` and then find common dtypes for the columns appeared to be in
            an intersection. This implements the logic of how dtypes are combined on
            ``pd.concat([df1, df2], axis=0).dtypes``.

        Returns
        -------
        ModinDtypes
        """
        preprocessed_vals = []
        for val in values:
            if isinstance(val, cls):
                val = val._value
            if isinstance(val, (DtypesDescriptor, pandas.Series)) or val is None:
                preprocessed_vals.append(val)
            else:
                raise NotImplementedError(type(val))

        try:
            desc = DtypesDescriptor.concat(preprocessed_vals, axis=axis)
        except NotImplementedError as e:
            # 'DtypesDescriptor' doesn't support duplicated labels, however, if all values are pandas Series,
            # we still can perform concatenation using pure pandas
            if (
                # 'pd.concat(axis=1)' fails on duplicated labels anyway, so doing this logic
                # only in case 'axis=0'
                axis == 0
                and "duplicated" not in e.args[0].lower()
                or not all(isinstance(val, pandas.Series) for val in values)
            ):
                raise e
            desc = pandas.concat(values)
        return ModinDtypes(desc)

    def set_index(self, new_index: Union[pandas.Index, "ModinIndex"]) -> "ModinDtypes":
        """
        Set new column names for stored dtypes.

        Parameters
        ----------
        new_index : pandas.Index or ModinIndex

        Returns
        -------
        ModinDtypes
            New ``ModinDtypes`` with updated column names.
        """
        new_self = self.copy()
        if self.is_materialized:
            new_self._value.index = new_index
            return new_self
        elif callable(self._value):
            old_val = new_self._value
            new_self._value = lambda: old_val().set_axis(new_index)
            return new_self
        elif isinstance(new_self._value, DtypesDescriptor):
            new_self._value = new_self._value.set_index(new_index)
            return new_self
        else:
            raise NotImplementedError()

    def get(self) -> pandas.Series:
        """
        Get the materialized internal representation.

        Returns
        -------
        pandas.Series
        """
        if not self.is_materialized:
            if callable(self._value):
                self._value = self._value()
                if self._value is None:
                    self._value = pandas.Series([])
            elif isinstance(self._value, DtypesDescriptor):
                self._value = self._value.to_series()
            else:
                raise NotImplementedError(type(self._value))
        return self._value

    def __len__(self):
        """
        Redirect the 'len' request to the internal representation.

        Returns
        -------
        int

        Notes
        -----
        Executing this function materializes the data.
        """
        if not self.is_materialized:
            self.get()
        return len(self._value)

    def __reduce__(self):
        """
        Serialize an object of this class.

        Returns
        -------
        tuple

        Notes
        -----
        The default implementation generates a recursion error. In a short:
        during the construction of the object, `__getattr__` function is called, which
        is not intended to be used in situations where the object is not initialized.
        """
        return (self.__class__, (self._value,))

    def __getattr__(self, name):
        """
        Redirect access to non-existent attributes to the internal representation.

        This is necessary so that objects of this class in most cases mimic the behavior
        of the ``pandas.Series``. The main limitations of the current approach are type
        checking and the use of this object where pandas dtypes are supposed to be used.

        Parameters
        ----------
        name : str
            Attribute name.

        Returns
        -------
        object
            Attribute.

        Notes
        -----
        Executing this function materializes the data.
        """
        if not self.is_materialized:
            self.get()
        return self._value.__getattribute__(name)

    def copy(self) -> "ModinDtypes":
        """
        Copy an object without materializing the internal representation.

        Returns
        -------
        ModinDtypes
        """
        idx_cache = self._value
        if not callable(idx_cache):
            idx_cache = idx_cache.copy()
        return ModinDtypes(idx_cache)

    def __getitem__(self, key):  # noqa: GL08
        if not self.is_materialized:
            self.get()
        return self._value.__getitem__(key)

    def __setitem__(self, key, item):  # noqa: GL08
        if not self.is_materialized:
            self.get()
        self._value.__setitem__(key, item)

    def __iter__(self):  # noqa: GL08
        if not self.is_materialized:
            self.get()
        return iter(self._value)

    def __contains__(self, key):  # noqa: GL08
        if not self.is_materialized:
            self.get()
        return key in self._value


class LazyProxyCategoricalDtype(pandas.CategoricalDtype):
    """
    A lazy proxy representing ``pandas.CategoricalDtype``.

    Parameters
    ----------
    categories : list-like, optional
    ordered : bool, default: False

    Notes
    -----
    Important note! One shouldn't use the class' constructor to instantiate a proxy instance,
    it's intended only for compatibility purposes! In order to create a new proxy instance
    use the appropriate class method `._build_proxy(...)`.
    """

    def __init__(self, categories=None, ordered=False):
        # These will be initialized later inside of the `._build_proxy()` method
        self._parent, self._column_name, self._categories_val, self._materializer = (
            None,
            None,
            None,
            None,
        )
        super().__init__(categories, ordered)

    @staticmethod
    def update_dtypes(dtypes, new_parent):
        """
        Update a parent for categorical proxies in a dtype object.

        Parameters
        ----------
        dtypes : dict-like
            A dict-like object describing dtypes. The method will walk through every dtype
            an update parents for categorical proxies inplace.
        new_parent : object
        """
        for key, value in dtypes.items():
            if isinstance(value, LazyProxyCategoricalDtype):
                dtypes[key] = value._update_proxy(new_parent, column_name=key)

    def _update_proxy(self, parent, column_name):
        """
        Create a new proxy, if either parent or column name are different.

        Parameters
        ----------
        parent : object
            Source object to extract categories on demand.
        column_name : str
            Column name of the categorical column in the source object.

        Returns
        -------
        pandas.CategoricalDtype or LazyProxyCategoricalDtype
        """
        if self._is_materialized:
            # The parent has been materialized, we don't need a proxy anymore.
            return pandas.CategoricalDtype(self.categories, ordered=self._ordered)
        elif parent is self._parent and column_name == self._column_name:
            return self
        else:
            return self._build_proxy(parent, column_name, self._materializer)

    @classmethod
    def _build_proxy(cls, parent, column_name, materializer, dtype=None):
        """
        Construct a lazy proxy.

        Parameters
        ----------
        parent : object
            Source object to extract categories on demand.
        column_name : str
            Column name of the categorical column in the source object.
        materializer : callable(parent, column_name) -> pandas.CategoricalDtype
            A function to call in order to extract categorical values.
        dtype : dtype, optional
            The categories dtype.

        Returns
        -------
        LazyProxyCategoricalDtype
        """
        result = cls()
        result._parent = parent
        result._column_name = column_name
        result._materializer = materializer
        result._dtype = dtype
        return result

    def _get_dtype(self):
        """
        Get the categories dtype.

        Returns
        -------
        dtype
        """
        if self._dtype is None:
            self._dtype = self.categories.dtype
        return self._dtype

    def __reduce__(self):
        """
        Serialize an object of this class.

        Returns
        -------
        tuple

        Notes
        -----
        This object is serialized into a ``pandas.CategoricalDtype`` as an actual proxy can't be
        properly serialized because of the references it stores for its potentially distributed parent.
        """
        return (pandas.CategoricalDtype, (self.categories, self.ordered))

    @property
    def _categories(self):
        """
        Get materialized categorical values.

        Returns
        -------
        pandas.Index
        """
        if not self._is_materialized:
            self._materialize_categories()
        return self._categories_val

    @_categories.setter
    def _categories(self, categories):
        """
        Set new categorical values.

        Parameters
        ----------
        categories : list-like
        """
        self._categories_val = categories
        self._parent = None  # The parent is not required any more
        self._materializer = None
        self._dtype = None

    @property
    def _is_materialized(self) -> bool:
        """
        Check whether categorical values were already materialized.

        Returns
        -------
        bool
        """
        return self._categories_val is not None

    def _materialize_categories(self):
        """Materialize actual categorical values."""
        ErrorMessage.catch_bugs_and_request_email(
            failure_condition=self._parent is None,
            extra_log="attempted to materialize categories with parent being 'None'",
        )
        categoricals = self._materializer(self._parent, self._column_name)
        self._categories = categoricals.categories
        self._ordered = categoricals.ordered


def get_categories_dtype(
    cdt: Union[LazyProxyCategoricalDtype, pandas.CategoricalDtype]
):
    """
    Get the categories dtype.

    Parameters
    ----------
    cdt : LazyProxyCategoricalDtype or pandas.CategoricalDtype

    Returns
    -------
    dtype
    """
    return (
        cdt._get_dtype()
        if isinstance(cdt, LazyProxyCategoricalDtype)
        else cdt.categories.dtype
    )


def extract_dtype(value):
    """
    Extract dtype(s) from the passed `value`.

    Parameters
    ----------
    value : object

    Returns
    -------
    numpy.dtype or pandas.Series of numpy.dtypes
    """
    from modin.pandas.utils import is_scalar

    if hasattr(value, "dtype"):
        return value.dtype
    elif hasattr(value, "dtypes"):
        return value.dtypes
    elif is_scalar(value):
        return np.dtype(type(value))
    else:
        return np.array(value).dtype



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

"""Dataframe exchange protocol tests that are specific for pandas storage format implementation."""

import pandas

import modin.pandas as pd
from modin.pandas.test.utils import df_equals, test_data
from modin.pandas.utils import from_dataframe
from modin.test.test_utils import warns_that_defaulting_to_pandas


def eval_df_protocol(modin_df_producer):
    internal_modin_df_producer = modin_df_producer.__dataframe__()
    # Our configuration in pytest.ini requires that we explicitly catch all
    # instances of defaulting to pandas, this one raises a warning on `.from_dataframe`
    with warns_that_defaulting_to_pandas():
        modin_df_consumer = from_dataframe(modin_df_producer)
        internal_modin_df_consumer = from_dataframe(internal_modin_df_producer)

    # TODO: the following assertions verify that `from_dataframe` doesn't return
    # the same object untouched due to optimization branching, it actually should
    # do so but the logic is not implemented yet, so the assertions are passing
    # for now. It's required to replace the producer's type with a different one
    # to consumer when we have some other implementation of the protocol as the
    # assertions may start failing shortly.
    assert modin_df_producer is not modin_df_consumer
    assert internal_modin_df_producer is not internal_modin_df_consumer
    assert (
        modin_df_producer._query_compiler._modin_frame
        is not modin_df_consumer._query_compiler._modin_frame
    )

    df_equals(modin_df_producer, modin_df_consumer)
    df_equals(modin_df_producer, internal_modin_df_consumer)


def test_simple_import():
    modin_df = pd.DataFrame(test_data["int_data"])
    eval_df_protocol(modin_df)


def test_categorical_from_dataframe():
    modin_df = pd.DataFrame(
        {"foo": pd.Series(["0", "1", "2", "3", "0", "3", "2", "3"], dtype="category")}
    )
    eval_df_protocol(modin_df)


def test_from_dataframe_with_empty_dataframe():
    modin_df = pd.DataFrame({"foo_col": pd.Series([], dtype="int64")})
    with warns_that_defaulting_to_pandas():
        eval_df_protocol(modin_df)


def test_interchange_with_pandas_string():
    modin_df = pd.DataFrame({"fips": ["01001"]})
    pandas_df = pandas.api.interchange.from_dataframe(modin_df.__dataframe__())
    df_equals(modin_df, pandas_df)



# Licensed to Modin Development Team under one or more contributor license agreements.
# See the NOTICE file distributed with this work for additional information regarding
# copyright ownership.  The Modin Development Team licenses this file to you under the
# Apache License, Version 2.0 (the "License"); you may not use this file except in
# compliance with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.


"""Collection of utility functions for the PandasDataFrame."""

import pandas
from pandas.api.types import union_categoricals

from modin.error_message import ErrorMessage


def concatenate(dfs, copy=True):
    """
    Concatenate pandas DataFrames with saving 'category' dtype.

    All dataframes' columns must be equal to each other.

    Parameters
    ----------
    dfs : list
        List of pandas DataFrames to concatenate.
    copy : bool, default: True
        Make explicit copy when creating dataframe.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame.
    """
    for df in dfs:
        assert df.columns.equals(dfs[0].columns)
    for i in dfs[0].columns.get_indexer_for(dfs[0].select_dtypes("category").columns):
        columns = [df.iloc[:, i] for df in dfs]
        all_categorical_parts_are_empty = None
        has_non_categorical_parts = False
        for col in columns:
            if isinstance(col.dtype, pandas.CategoricalDtype):
                if all_categorical_parts_are_empty is None:
                    all_categorical_parts_are_empty = len(col) == 0
                    continue
                all_categorical_parts_are_empty &= len(col) == 0
            else:
                has_non_categorical_parts = True
        # 'union_categoricals' raises an error if some of the passed values don't have categorical dtype,
        # if it happens, we only want to continue when all parts with categorical dtypes are actually empty.
        # This can happen if there were an aggregation that discards categorical dtypes and that aggregation
        # doesn't properly do so for empty partitions
        if has_non_categorical_parts and all_categorical_parts_are_empty:
            continue
        union = union_categoricals(columns)
        for df in dfs:
            df.isetitem(
                i, pandas.Categorical(df.iloc[:, i], categories=union.categories)
            )
    # `ValueError: buffer source array is read-only` if copy==False
    if len(dfs) == 1 and copy:
        # concat doesn't make a copy if len(dfs) == 1,
        # so do it explicitly
        return dfs[0].copy()
    return pandas.concat(dfs, copy=copy)


def create_pandas_df_from_partitions(
    partition_data, partition_shape, called_from_remote=False
):
    """
    Convert partition data of multiple dataframes to a single dataframe.

    Parameters
    ----------
    partition_data : list
        List of pandas DataFrames or list of Object references holding pandas DataFrames.
    partition_shape : int or tuple
        Shape of the partitions NumPy array.
    called_from_remote : bool, default: False
        Flag used to check if explicit copy should be done in concat.

    Returns
    -------
    pandas.DataFrame
        A pandas DataFrame.
    """
    if all(
        isinstance(obj, (pandas.DataFrame, pandas.Series)) for obj in partition_data
    ):
        height, width, *_ = tuple(partition_shape) + (0,)
        # restore 2d array
        objs = iter(partition_data)
        partition_data = [[next(objs) for _ in range(width)] for __ in range(height)]
    else:
        # Partitions do not always contain pandas objects, for example, hdk uses pyarrow tables.
        # This implementation comes from the fact that calling `partition.get`
        # function is not always equivalent to `partition.to_pandas`.
        partition_data = [[obj.to_pandas() for obj in part] for part in partition_data]
    if all(isinstance(part, pandas.Series) for row in partition_data for part in row):
        axis = 0
    elif all(
        isinstance(part, pandas.DataFrame) for row in partition_data for part in row
    ):
        axis = 1
    else:
        ErrorMessage.catch_bugs_and_request_email(True)

    def is_part_empty(part):
        return part.empty and (
            not isinstance(part, pandas.DataFrame) or (len(part.columns) == 0)
        )

    df_rows = [
        pandas.concat([part for part in row], axis=axis, copy=False)
        for row in partition_data
        if not all(is_part_empty(part) for part in row)
    ]
    copy = not called_from_remote
    if len(df_rows) == 0:
        return pandas.DataFrame()
    else:
        return concatenate(df_rows, copy)
